/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2022, Kassow Robots
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Kassow Robots nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

#include <kr2/utils/utils.h>
#include <kr2/utils/extrapolator.h>
#include <kr2/utils/pyremoter.h>
#include <kr2/utils/uploader.h>
#include <kr2/utils/heartbeat_monitor.h>
#include <kr2/utils/commands_monitor.h>
#include <kr2/utils/configuration_file.h>
#include <kr2/utils/cpu_reader.h>
#include <kr2/kord/robot_control_action.h>
#include <kr2/kord/robot_model.h>
#include <kr2/kord/robot_controller.h>
#include <kr2/kord/udp_kord.h>
#include <kr2/kord/kord_server.h>
#include <kr2/kord/protocol/JointFirmwareCommand.h>

#include <kr2_rc_api/api_v1/arm/model.h>
#include "kr2_rc_api/api_v1/io/iodata.h"
#include <kr2_rc_api/api_v1/arm/trajectory.h>
#include <kr2_rc_api/internals/timex.h>

#include <kr2_program_api/api_v1/bundles/arg_provider_xml.h>
#include <boost/algorithm/string/predicate.hpp>

#include <ctime>
#include <thread>
#include <bitset>

#include <chrono>
#include <regex>
#include <kr2/utils/stats.h>

#ifndef GIT_VERSION
    #include <kr2/version.h>
#endif

#include <kr2/kord/system/SystemAlarm.h>
#include <kr2/kord/system/RobotControllerFlags.h>

using namespace kswx_kr2_kord;
using namespace kr2::kord;
using namespace kr2::kord::protocol;

REGISTER_CLASS(kswx_kr2_kord::UDPKord)
using namespace kr2rc_api_internals;

class KORDReceiver
{
public:
    KORDReceiver(boost::shared_ptr<kr2_program_api::ProgramInterface> a_api, int a_port = 7582, bool a_sync_w_rc = true)
        : keep_running_(false),
          program_api_(a_api),
          extr_ws_(kord_period_, 6),
          extr_js_(kord_period_, 7),
          control_timeout_(std::chrono::microseconds(50000))
    {
        kr2rc_api_internals::Timex::initSingletonTimex(500, false);
    }

    ~KORDReceiver()
    {
    }

    int run();

    void printStats(CommandsMonitor& stats_commands);

    void processControlActions(const RobotControl &controls, Uploader &u, HeartBeatMonitor &stats_monitior,
    CommandsMonitor &stats_commands,
    kr2::kord::KORDServer & k_server, bool recent = true);

    void doCBunError(std::string a_description);

    void processPendingStatuses();

    void processTransferStatus(Uploader &u);

    int start(int a_port = 7582, int a_session_id = 1, bool a_sync_w_rc = true, bool a_use_rt = false, int a_rt_prio = 1, 
    std::string a_config_file_path = "/home/nucky/kr2.root/Workspace/Files/")
    {
        port_ = a_port;
        session_id_ = a_session_id;
        sync_w_rc_ = a_sync_w_rc;
        use_rt_ = a_use_rt;
        rt_prio_ = a_rt_prio;

        //reading the config file
        config_file_.chooseFile(a_config_file_path);
        config_file_.read();

        if (keep_running_.load())
        {
            return -1;
        }

        keep_running_.store(true);
        // std::string SID = "[KORD_CBUN] Session id" + std::string(session_id_);
        CLOG_INFO("[KORD_CBUN] Session id: " << session_id_);

        // API Parameters
        /*
         * The combination of writing the move command and reading Frames and Loads
         * caused the syncModel() to use on average 4ms of time, which as a result 
         * extended the runtime to expand over to another tick and sleep for two 
         * updates. This effectively made the heartbeat arrive at rate of 125Hz. Result
         * of which was the streamed control commands arrived at rate of 125Hz and the 
         * arm failed to move at all since the commands appeared to time out.
         */
        program_api_->rc_api_->disableAPIFeatures(std::set<kr2rc_api::APIParameters::EFeatures>{
            kr2rc_api::APIParameters::EFeatures::ARM_MODEL_SYNC_USER_READS,
            kr2rc_api::APIParameters::EFeatures::ARM_LOAD_SYNC_USER_READS
        });
        CLOG_INFO("[KORD_CBUN] Receiver start");
        receiver_ = std::thread([this]()
                                { run(); });
        CLOG_INFO("[KORD_CBUN] Receiver started");

        //cpu reader thread start (only if enabled)
        if (config_file_.get_data().CPU_READER_ENABLE){
            cpu_reader_.start();
        }

        return 0;
    }

    int stop()
    {
        if (keep_running_.load())
        {
            keep_running_.store(false);

            // The realtime thread is detached from the main and is not joinable.
            if (use_rt_)
                return 0;

            if (receiver_.joinable())
            {
                receiver_.join();
            }

        }

        return 0;
    }

private:
    // Blocks untile next update and then reads positions of joints
    //    void syncRCData(kr2::kord::KR2KORDFrame_StatusJ_Res *a_jstate);

    // Saves Dashboard Json Configuration file to the specified location
    bool askForDashboardJsonToFile();

    // Blocks until the next tick starts, next tick from provided time stamp
    void syncTick(const timespec &ts, CommandsMonitor&);
    // Blocks until the next tick starts, reads current time stamp
    void syncTick(int nticks, CommandsMonitor&, timespec a_tick_start ={0,0});

    // Compose systemAlarmState
    uint32_t systemAlarmState(CommandsMonitor& stats_commands);

    // Spins RC API and stores read values in a local arm model.
    bool syncModel(CommandsMonitor&);

    // Tracking command in joint/work spaces
    template <class A, size_t B>
    void sendTTCommand(const std::array<A, B> &a_target, const kr2rc_api::Trajectory::CmdTrajectoryParams &params);

    // Tracking command in Direct Movement
    void sendTTCommandD(
        const std::array<double, 7UL> &a_q,
        const std::array<double, 7UL> &a_qd,
        const std::array<double, 7UL> &a_qdd,
        const std::array<double, 7UL> &a_torque);

    // Velocity movement
    void sendTTCommandV(
        const std::array<double, 6UL> &a_v,
        const long a_sync_value,
        const double a_period,
        const double a_timeout);
    // Set Frame command
    kr2rc_api::CmdResult sendTTCommandSetFrame(
        unsigned int frame_id,
        const std::array<double, 6UL> &a_p,
        unsigned int ref_frame,
        unsigned int& executed_token);

    // Set Load command
    kr2rc_api::CmdResult sendTTCommandSetLoad(
        unsigned int a_load_id,
        const std::array<double, 6UL> &a_inertia,
        double a_mass,
        const std::array<double, 3UL> &a_cog,
        unsigned int& executed_token);

    // Set Clean Alarm command
    kr2rc_api::CmdResult sendTTCommandCleanAlarm(
        unsigned int a_alarm_id,
        unsigned int& executed_token, 
        CommandsMonitor& stats_commands);

    void sendSetIOCommand(const RobotControl &a_controls);

    kr2rc_api::CmdResult sendDJC_CommandFW(unsigned int, EJointFirmwareCommand, unsigned int& exec_token_);

    // Handle processing of firmware commands
    // @pre     RobotControl needs to hols a valid timestamp of the last command send.
    // @post    RobotControl firmware command time stamp is updated if there are commands
    //          to be transmitted. The commands are droppped from the command buffer.
    void processFirmwareCommands(RobotControl &controls);

    bool readModelTF(std::vector<double> &a_out, kr2rc_api::DUID a_duid);

    bool readModelTFWithRef(std::vector<double> &a_out, kr2rc_api::DUID a_duid, kr2rc_api::DUID a_ref_duid);

    bool readModelLoadCoG(std::vector<double> &a_out, kr2rc_api::DUID a_duid);
    bool readModelLoadInertia(std::vector<double> &a_out, kr2rc_api::DUID a_duid);
    bool readModelLoadPose(std::vector<double> &a_out, kr2rc_api::DUID a_duid);
    bool readModelLoadMass(double &a_out, kr2rc_api::DUID a_duid);

    bool readModelJoints(std::vector<double> &a_out, kr2rc_api::DUID a_duid);

    bool readJointSensors(JointSensorData &sensors, unsigned int idx);

    bool readIOBoardSensors(IOBoardData &a_ioboard);

    bool readPeripherals();

    bool readDigital(long, std::vector<kr2rc_api::IOData::GPIOInt64> &);

    bool readAnalog(long, std::vector<kr2rc_api::IOData::GPIOFloat> &);

    /**
     * @brief Reads safety events and soft stop events from the RC API.
     *
     * @param events Circular buffer storing latest 10 events.
     * @return true if the events were collected successfully
     * @return false if the events could not be collected
     */
    bool readSystemEvents(SystemEvents &events);

private:
    int port_;
    int session_id_;
    std::string ifname_;
    std::atomic<bool> keep_running_;
    std::thread receiver_;

    CPUStateReader cpu_reader_ = CPUStateReader();
    ConfifurationFile config_file_ = ConfifurationFile();

    boost::shared_ptr<kr2_program_api::ProgramInterface> program_api_;

    bool sync_w_rc_;

    bool use_rt_;
    int rt_prio_;

    // Control references.
    kr2::kord::RobotControl robot_control_;

    // State of the arm and cabinet.
    RobotModel arm_model_;
    // State of the SW controller (responsive controller)
    RobotController controller_model_;

    kr2::utils::VectorExtrapolator extr_ws_;
    kr2::utils::VectorExtrapolator extr_js_;

    std::chrono::microseconds control_timeout_;

    static constexpr std::chrono::milliseconds CMD_FW_DELAY = std::chrono::milliseconds(50);

    StatusBuffer<CommandStatus> command_statuses_;

    const int MAX_CHECK_TIMES_FOR_PENDING = 20;

    StatusBuffer<PendingStatus> pending_statuses_;

    const unsigned int TOKEN_MAX_ = std::numeric_limits<unsigned int>::max();
    unsigned int last_token_ = 0;

    std::map<int64_t, int> accum_pending_results; // number of successful pending results got for each token

    int accum_pending_result = 0; // number of successful pending results got

    int64_t last_sent_timestamp_ = -1;

    static constexpr unsigned int kord_frequency_ = 250;
    static constexpr double kord_period_ = 1.0 / kord_frequency_;
    static constexpr unsigned int kord_period_micros = 4000;

    SystemEvent recent_safety_event_;
    SystemEvent recent_soft_stop_event_;


    Accum<int64_t> fetch_times = Accum<int64_t>(10000);
    int64_t global_fetch_times_max = 0;
    Accum<int64_t> dispatch_times = Accum<int64_t>(10000);
    int64_t global_dispatch_times_max = 0;
    Accum<int64_t> sync_model_times = Accum<int64_t>(10000);
    int64_t global_sync_model_times_max = 0;
};

void printHorizontalLine(int len) {
    std::cout << std::setw(len) << std::setfill('-') << "";
    std::cout << std::setfill(' ') << std::endl;
}

void KORDReceiver::printStats(CommandsMonitor& stats_commands){
    const int columns = 5;
    const int columnWidth = 10;
    const int firstColumnWidth = 20;

    // Print header
    printHorizontalLine((columns - 1) * columnWidth + firstColumnWidth);
    std::cout << std::left << std::setw(firstColumnWidth) << "Metrics (micros)"
              << "|" << std::setw(columnWidth - 1) << "GlobalMax"
              << "|" << std::setw(columnWidth - 1) << "WindMin"
              << "|" << std::setw(columnWidth - 1) << "WindMax"
              << "|" << std::setw(columnWidth - 1) << "WindAvg" << "|" << std::endl;
    printHorizontalLine((columns - 1) * columnWidth + firstColumnWidth);

    // Print data
    std::cout << std::left << std::setw(firstColumnWidth) << "|Sync Wake-Up Jitter"
        << "|" << std::setw(columnWidth - 1) << stats_commands.global_system_jitter_max
        << "|" << std::setw(columnWidth - 1) << stats_commands.accum_wake_up_stats.get_min()
        << "|" << std::setw(columnWidth - 1) << stats_commands.accum_wake_up_stats.get_max()
        << "|" << std::setw(columnWidth - 1) << stats_commands.accum_wake_up_stats.get_avg() << "|" << std::endl;
    std::cout << std::left << std::setw(firstColumnWidth) << "|Fetch takes" 
        << "|" << std::setw(columnWidth - 1) << global_fetch_times_max /1000.
        << "|" << std::setw(columnWidth - 1) << fetch_times.get_min() /1000.
        << "|" << std::setw(columnWidth - 1) << fetch_times.get_max() /1000.
        << "|" << std::setw(columnWidth - 1) << fetch_times.get_avg() /1000. << "|" << std::endl;
    std::cout << std::left << std::setw(firstColumnWidth) << "|Dispatch takes"
        << "|" << std::setw(columnWidth - 1) << global_dispatch_times_max/1000.
        << "|" << std::setw(columnWidth - 1) << dispatch_times.get_min() /1000.
        << "|" << std::setw(columnWidth - 1) << dispatch_times.get_max() /1000.
        << "|" << std::setw(columnWidth - 1) << dispatch_times.get_avg() /1000. << "|" << std::endl;
    std::cout << std::left << std::setw(firstColumnWidth) << "|Sync Model takes"  
        << "|" << std::setw(columnWidth - 1) << global_sync_model_times_max/1000.
        << "|" << std::setw(columnWidth - 1) << sync_model_times.get_min() /1000.
        << "|" << std::setw(columnWidth - 1) << sync_model_times.get_max() /1000.
        << "|" << std::setw(columnWidth - 1) << sync_model_times.get_avg() /1000. << "|" << std::endl;
    printHorizontalLine((columns - 1) * columnWidth + firstColumnWidth);
}

int KORDReceiver::run()
{
    kr2::kord::KORDServer kord_server(port_, session_id_, kr2::kord::EConnType::KORD_UDP_CONN, kord_frequency_);
    auto start = std::chrono::high_resolution_clock::now();
    unsigned long late_data_cnt = 0;
    unsigned long recent_data_cnt = 0;
    arm_model_.reset();

    if (use_rt_)
    {
        kr2::utils::init_realtime_params(rt_prio_);
    }

    // Open connection by creating the socket and starting the capture thread
    CLOG_INFO("[KORD_CBUN] Transmitter start");
    if (!kord_server.open(10.0))
    {
        CLOG_ERR("[CBUN_KORD] Failed to start server");
        return -1;
    }

    CLOG_INFO("[KORD_CBUN] Transmitter started");

    CLOG_INFO(std::to_string(config_file_.get_data().HALT_TRIGGER));

    CommandsMonitor stats_commands = CommandsMonitor(program_api_, config_file_.get_data());

    // sync to RC control loop
    syncTick(2, stats_commands);

    // update model for first dispatch
    syncModel(stats_commands);
    CLOG_INFO("[KORD_CBUN] First sync OK");

    kord_server.updateStatus(arm_model_, controller_model_, command_statuses_);

    // main loop
    Uploader u = Uploader();
    u.set_config(config_file_.get_data());
    u.set_program_api(program_api_);
    HeartBeatMonitor stats_monitor = HeartBeatMonitor(); // <- status frame echoes not used
    CLOG_INFO("[KORD_CBUN] Main loop starting, ts: " << start.time_since_epoch().count());
    int64_t stats_diff;
    while (keep_running_.load())
    {
        controller_model_.system_events_.reset(); // cleaning the buffer before syncing
        timespec now_ts = kr2rc_api_internals::Timex::now();
        /*
         * Tick time is initialized to 500Hz to allow reading the RC API
         * and writing it properly. If the frequency is mismatched, the data
         * can get lost or reading can fail.
         *
         * Valid captured frames are -2 for 4ms and -4 to 8ms to allow more give.
         *
         */
        tick_t now_tick = kr2rc_api_internals::Timex::thisTick(now_ts, -2);

        // check valid recent commands available
        bool recent_data = kord_server.fetchItem(&robot_control_, now_tick, stats_commands, controller_model_);

        timespec fetch_ts = kr2rc_api_internals::Timex::now();
        stats_diff = kr2rc_api_internals::Timex::nsecDiff(fetch_ts, now_ts);
        fetch_times.measurement(stats_diff);
        if (stats_diff > global_fetch_times_max){
            global_fetch_times_max = stats_diff;
        }

        // time
        // transmit arm status
        kord_server.dispatch(last_sent_timestamp_, stats_commands);

        timespec dispatch_ts = kr2rc_api_internals::Timex::now();
        stats_diff = kr2rc_api_internals::Timex::nsecDiff(dispatch_ts, fetch_ts);
        dispatch_times.measurement(stats_diff);
        if (stats_diff > global_dispatch_times_max){
            global_dispatch_times_max = stats_diff;
        }

        // kord_server.computeCaptureThreadStats(0);
        // process the data
        processControlActions(robot_control_, u, stats_monitor, stats_commands, kord_server, recent_data);
        // (Processing just before dispatch to have time to fetch status echo)

        //check and update pending commands statuses
        processPendingStatuses();

        // Currently the firmware commands cannot be processed in a multiple manner and must be processed
        // sequentially. Therefor even if the capture time is old, it is still needed to execute the
        // firmware command until there are none.
        // The available commands regard the brakes. The brakes in use today (September 2022) do drain quite a lot
        // of current. For that reason it is advised to attempt brake release within certain time intervals. Therefore
        // a delay of 50ms between each command is introduced. Releasing all brakes takes ~350ms.
        processFirmwareCommands(robot_control_);

        // Processing the transfer status
        // ! WARNING ! It changes:
        // *    controller_model_
        processTransferStatus(u);
        // ! WARNING ! It changes:
        // *    controller_model_
        timespec before_syncModel_ts = kr2rc_api_internals::Timex::now();
        syncModel(stats_commands);
        stats_diff = kr2rc_api_internals::Timex::nsecDiff(kr2rc_api_internals::Timex::now(), before_syncModel_ts);
        sync_model_times.measurement(stats_diff);
        if (stats_diff > global_sync_model_times_max){
            global_sync_model_times_max = stats_diff;
        }

        //stats_commands.setSystemEventState(controller_model_.motion_flags_);

        kord_server.updateStatus(arm_model_, controller_model_, command_statuses_);

        // nanostleep until next tick
        // syncTick(now_ts);
        // currently syncing to 4ms (base freq of 500Hz means a base tick of 2ms)
        // 250Hz uses 2 base ticks
        syncTick(2, stats_commands);

        // if (stats_commands.sys_problem){
        //     fetch_times.compute();
        //     dispatch_times.compute();
        //     sync_model_times.compute();
        //     printStats(stats_commands);
        //     std::cout << '\n' << kr2rc_api_internals::Timex::nsecDiff(fetch_ts, now_ts) << " "
        //               << kr2rc_api_internals::Timex::nsecDiff(dispatch_ts, fetch_ts) << " "
        //               << stats_diff << std::flush;
        //     stats_commands.sys_problem = false;
        // }
    }

    CLOG_INFO("[CBUN_KORD] Loop finished, closing server");
    if (kord_server.close())
    {
        CLOG_WARN("[CBUN_KORD] Failed to close the server");
    }
    u.join();

    stats_monitor.reset();

    cpu_reader_.stop();
    cpu_reader_.join();

    //stats_commands.accum_wakeUp_stats.compute();
    
    fetch_times.compute();
    dispatch_times.compute();
    sync_model_times.compute();

    printStats(stats_commands);
    std::cout << std::flush;

    return 0;
}

/// @brief Processing TransferStatus from PyRemoter
/// @param u
/// ! WARNING ! It changes:
/// *   controller_model_
void KORDReceiver::processTransferStatus(Uploader &u)
{
    // if (controller_model_.req_sys_tx == 0) { return; }
    // if (a_controls.req_sys_tx == 0)        { return; }
    // todo: consider to fix the "2nd ok attempt"
    // if (controller_model_.req_sys_tx == robot_control_.req_sys_tx) {
    //     return;
    // }

    switch (robot_control_.req_system_id_)
    {
    case EControlCommandItems::eTransferLogFiles:
    case EControlCommandItems::eTransferDashboardJson:    
    case EControlCommandItems::eTransferCalibrationData:
    case EControlCommandItems::eTransferFiles:
    {
        auto status_transfer = u.get_status();
        switch (status_transfer.result_)
        {
        case kr2_program_api::CmdResult<>::ERROR:
        {
            controller_model_.request_system_status_ = EControlCommandStatus::eFailure;
            controller_model_.request_error_ = status_transfer.code_;
            u.reset();
            break;
        }
        case kr2_program_api::CmdResult<>::OK:
        {
            controller_model_.request_system_status_ = EControlCommandStatus::eSuccess;
            u.reset();
            break;
        }

        // ! The status of the Uploader is Exception[code:7, message:PYREMOTER_STATUS_SDESC[7]] -->  PYREMOTER_STATUS_SDESC[7] == "In the progress."
        case kr2_program_api::CmdResult<>::EXCEPTION:
        {
            controller_model_.req_sys_tx = robot_control_.req_sys_tx;
            controller_model_.req_system_id_ = robot_control_.req_system_id_;
            controller_model_.request_system_status_ = EControlCommandStatus::eInProgress;
            break;
        }
        }
        break;
    }

    case EControlCommandItems::eIOSet:
    {
        controller_model_.req_system_id_ = robot_control_.req_system_id_;
        controller_model_.req_sys_tx = robot_control_.req_sys_tx;
        controller_model_.request_system_status_ = EControlCommandStatus::eSuccess;
        break;
    }
    }
}

void KORDReceiver::processPendingStatuses(){
    long sys_state_n_ = program_api_->rc_api_->system_state_->read_SysStat_NReceipts();

    for (auto stat_ = pending_statuses_.statuses_.begin(); stat_ != pending_statuses_.statuses_.end();){
        bool now_found = false;
        for (int i = 0; i < sys_state_n_; ++i){
            auto new_receipt = program_api_->rc_api_->system_state_->read_SysStat_Receipt(i);
            if (new_receipt->value(kr2rc_api::State::SysStat_Receipt::EItem::UITM_TOKEN) == stat_->rc_token){
                now_found = true;
                // sending to kord-api = adding to command_statuses buffer
                if (stat_->composed){
                    CLOG_INFO("Found composed: " << accum_pending_results[stat_->token] << " " << stat_->compos_total - 1 << " " << stat_->compos_left);
                    if ((accum_pending_results[stat_->token] == stat_->compos_total - 1) && (stat_->compos_left == 0)){ // waiting all the dependent states
                        CommandStatus new_stat = CommandStatus();
                        new_stat.token = stat_->token;
                        new_stat.error_code = (new_receipt->value(kr2rc_api::State::SysStat_Receipt::EItem::UITM_SUCCESS) << 2) | 2; // Command error
                        command_statuses_.addStatus(new_stat);

                        CLOG_INFO("Removing token: " << new_stat.token);
                        accum_pending_result = 0;
                        accum_pending_results.erase(stat_->token);
                    }
                    else{
                    accum_pending_results[stat_->token] += 1; // waiting all the dependent states
                    }
                }
                else{ 
                        CommandStatus new_stat = CommandStatus();
                        new_stat.token = stat_->token;
                        new_stat.error_code = (new_receipt->value(kr2rc_api::State::SysStat_Receipt::EItem::UITM_SUCCESS) << 2) | 2; // Command error
                        command_statuses_.addStatus(new_stat);

                        CLOG_INFO("Removing token: " << new_stat.token);
                }
                
                // deleting from pending buffer
                stat_ = pending_statuses_.statuses_.erase(stat_);
            }
        }
        if (!now_found){
            stat_->times_checked += 1;
            CLOG_INFO("Updating rc-token: " << stat_->rc_token);

            if (stat_->times_checked > MAX_CHECK_TIMES_FOR_PENDING){
                // sending to kord-api = adding to command_statuses buffer

                CommandStatus new_stat = CommandStatus();
                new_stat.token = stat_->token;
                new_stat.error_code = (0 << 2) | 2; // Command error
                command_statuses_.addStatus(new_stat);

                if (stat_->composed){
                    accum_pending_result = 0;
                    accum_pending_results.erase(stat_->token);
                }
                
                // deleting from pending buffer
                stat_ = pending_statuses_.statuses_.erase(stat_);
                CLOG_INFO("Sending big not found: " << stat_->rc_token);
            }
            else{
                ++stat_;// Move to the next element
            }
        }
    }
}

void KORDReceiver::processControlActions(const RobotControl &a_controls, Uploader &u, HeartBeatMonitor & stats_monitor, 
CommandsMonitor &stats_commands, kr2::kord::KORDServer & k_server,
bool a_recent)
{

    // apply commands
    // send response
    // no relevant commands - extrapolate?
    // if movement was active update extrapolator
    // extrapolate max 5 steps - then EStop/torque deviation
    // else do nothing
    auto now_ts = std::chrono::high_resolution_clock::now();
    std::chrono::microseconds age = std::chrono::duration_cast<std::chrono::microseconds>(now_ts - a_controls.capture_ts_);

    if (age > control_timeout_)
    {
        return;
    }

    // early exit if any safety events
    SystemEvents __system_events;
    readSystemEvents(__system_events);
    switch (a_controls.item_id_)
    {
    case EKORDItemID::eCommandMoveJ:
    case EKORDItemID::eCommandMoveL:
    case EKORDItemID::eCommandMoveDirect:
    case EKORDItemID::eCommandMoveVelocity:
        // case EKORDItemID::eCommandCleanAlarm:
        if (!__system_events.is_empty())
        {
            CLOG_INFO("Safety Event is raised. Early exit.");
            return;
        }
        break;
    }
    // stats_commands.processNewSequenceNumber(a_controls.sequence_num_);

    // prefetch params for MoveL/MoveJ
    using TTP = kr2rc_api::Trajectory::CmdTrajectoryParams;
    kr2rc_api::Trajectory::CmdTrajectoryParams params;
    switch (a_controls.item_id_)
    {
    case EKORDItemID::eCommandMoveJ:
        params.tracking_type_ = a_controls.move_j_.getTrackingType();
        params.tracking_value_ = a_controls.move_j_.getTrackingTypeValue();
        params.blend_type_ = a_controls.move_j_.getBlendType();
        params.blend_value_ = a_controls.move_j_.getBlendTypeValue();
        params.overlay_type_ = a_controls.move_j_.getOverlayType();
        params.sync_time_ = a_controls.move_j_.getSyncTime();
        break;
    case EKORDItemID::eCommandMoveL:
        params.tracking_type_ = a_controls.move_l_.getTrackingType();
        params.tracking_value_ = a_controls.move_l_.getTrackingTypeValue();
        params.blend_type_ = a_controls.move_l_.getBlendType();
        params.blend_value_ = a_controls.move_l_.getBlendTypeValue();
        params.overlay_type_ = a_controls.move_l_.getOverlayType();
        params.sync_time_ = a_controls.move_l_.getSyncTime();
        break;
    }

    // Debug monitoring of actions
    // CLOG_INFO("[PROCESS_CTRL_ACTION]     age: " << age.count());
    // CLOG_INFO("[PROCESS_CTRL_ACTION] item_id: " << static_cast<unsigned int>(a_controls.item_id_));

    switch (a_controls.item_id_)
    {
    case EKORDItemID::eRequestStatusArm:
    {
        break;
    }
    case EKORDItemID::eCommandMoveJ:
    {
        if (a_recent)
        {
            std::array<double, 7UL> q = a_controls.move_j_.getTargets();
            std::vector<double> qv;
            std::copy(q.begin(), q.end(), std::back_inserter(qv));

            sendTTCommand<double, 7UL>(q, params);
            extr_js_.update(qv);
        }
        else
        {
            std::vector<double> qv = extr_js_.extrapolate();
            std::array<double, 7UL> q;
            std::copy_n(qv.begin(), q.size(), q.begin());
            if (qv.size() != 7)
            {
                CLOG_INFO("MOVE JOINTS: Vector size unexpected (should be 7), is: " << qv.size());
            }

            sendTTCommand<double, 7UL>(q, params);
            CLOG_WARN("[TS] " << now_ts.time_since_epoch().count() << " ns, [CBUN_KORD] Extrapolation JS active, steps extrapolated: " << extr_js_.counter());
        }
        break;
    }
    case EKORDItemID::eCommandMoveL:
    {
        if (a_recent)
        {
            std::array<double, 6UL> tcp = a_controls.move_l_.getTargets();
            std::vector<double> tcpv;
            std::copy(tcp.begin(), tcp.end(), std::back_inserter(tcpv));

            sendTTCommand<double, 6UL>(tcp, params);
            extr_ws_.update(tcpv);
        }
        else
        {
            std::vector<double> tcpv = extr_ws_.extrapolate();
            std::array<double, 6UL> tcp;
            std::copy_n(tcpv.begin(), tcp.size(), tcp.begin());
            if (tcpv.size() != 6)
            {
                CLOG_INFO("MOVE LINEAR: Vector size unexpected (should be 6), is: " << tcpv.size());
            }
            
            sendTTCommand<double, 6UL>(tcp, params);
            CLOG_WARN("[TS] " << now_ts.time_since_epoch().count() << " ns, [CBUN_KORD] Extrapolation WS active, steps extrapolated: " << extr_ws_.counter());
        }
        break;
    }
    case EKORDItemID::eStatusEchoResponse:
    {
        if (a_recent){
            // CLOG_INFO("Status Echo captured... ");
            // CLOG_INFO("    TimeStamp: " << a_controls.status_echo_tx);

            // CLOG_INFO("Processing Request... ");

            stats_monitor.processNewData(last_sent_timestamp_, a_controls.status_echo_tx, a_controls.status_echo_rx);
        }
        break;

    }
    case EKORDItemID::eRequestSystem:
    {
        CLOG_INFO("System Request captured... ");
        CLOG_INFO("    Request ID: " << a_controls.req_system_id_);
        CLOG_INFO("    Remote  IP: " << inet_ntoa(a_controls.remote_host_.sin_addr));
        CLOG_INFO("    CtrlMd  TX: " << controller_model_.req_sys_tx);
        CLOG_INFO("    Request TX: " << a_controls.req_sys_tx);

        if (controller_model_.req_sys_tx == a_controls.req_sys_tx)
        {
            break;
        }

        CLOG_INFO("Processing Request... ");

        switch (a_controls.req_system_id_)
        {
        case EControlCommandItems::eTransferLogFiles:
        {
            CLOG_INFO("    -> Transfer Logfiles request");
            // -pthread Lib is required
            // u.set_credfname("KORD.ini"); // default "KORD.ini"
            u.add_type(Uploader::LOGS_UPLOADER);
            u.set_host(inet_ntoa(a_controls.remote_host_.sin_addr));
            u.start();
            break;
        }
        case EControlCommandItems::eTransferDashboardJson:
        {
            CLOG_INFO("    -> Transfer Dashboard Json request");
            // -pthread Lib is required
            // u.set_credfname("KORD.ini"); // default "KORD.ini"

            if (askForDashboardJsonToFile()){
                u.add_type(Uploader::DIAGS_UPLOADER);
                u.set_host(inet_ntoa(a_controls.remote_host_.sin_addr));
                u.start();
            }
            break;
        }
        case EControlCommandItems::eTransferCalibrationData:
        {
            CLOG_INFO("    -> Transfer Calibration Data request");
            // -pthread Lib is required
            // u.set_credfname("KORD.ini"); // default "KORD.ini"
            u.add_type(Uploader::CALIB_UPLOADER);
            u.set_host(inet_ntoa(a_controls.remote_host_.sin_addr));
            u.start();
            break;
        }
        }
        break;
    }

    case EKORDItemID::eRequestTransfer:
    {
        CLOG_INFO("Transfer Request captured... ");
        CLOG_INFO("    Request ID: " << a_controls.req_system_id_);
        CLOG_INFO("    Remote  IP: " << inet_ntoa(a_controls.remote_host_.sin_addr));
        CLOG_INFO("    CtrlMd  TX: " << controller_model_.req_sys_tx);
        CLOG_INFO("    Request TX: " << a_controls.req_sys_tx);

        if (controller_model_.req_sys_tx == a_controls.req_sys_tx)
        {
            break;
        }

        CLOG_INFO("Processing Request... ");

        switch (a_controls.req_system_id_)
        {
        case EControlCommandItems::eTransferFiles:
        {
            CLOG_INFO("    -> Transfer Files request");
            // -pthread Lib is required
            // u.set_credfname("KORD.ini"); // default "KORD.ini"

            for (int type = 0; type < Uploader::UploaderType::SIZE; ++type){
                bool is_set = std::bitset<32>(a_controls.transfer_mask_).test(type);
                CLOG_INFO(a_controls.transfer_mask_ << " ; bit " << type << " is_set: " << is_set);
                if (is_set){
                    u.add_type(static_cast<Uploader::UploaderType>(type));
                }
            }

            u.set_host(inet_ntoa(a_controls.remote_host_.sin_addr));
            u.start();
            break;
        }
        default:
            CLOG_INFO("[Not covered system id: " << a_controls.req_system_id_ << "]");
        break;
        }

        break;
    }

    // case EKORDItemID::eCommandSetIOAnalogOut:
    case EKORDItemID::eCommandSetIODigitalOut:
    {
        CLOG_INFO("A/D IOSet Request captured... ");
        CLOG_INFO("    Request ID: " << a_controls.req_system_id_);
        CLOG_INFO("    Remote  IP: " << inet_ntoa(a_controls.remote_host_.sin_addr));
        CLOG_INFO("    CtrlMd  TX: " << controller_model_.req_sys_tx);
        CLOG_INFO("    Request TX: " << a_controls.req_sys_tx);

        if (controller_model_.req_sys_tx == a_controls.req_sys_tx)
        {
            break;
        }

        CLOG_INFO("Processing Request... ");

        switch (a_controls.req_system_id_)
        {
        case EControlCommandItems::eIOSet:
        {
            CLOG_INFO("-> Working on IOSet!");
            sendSetIOCommand(a_controls);
            break;
        }

        default:
            CLOG_INFO("[Not covered system id: " << a_controls.req_system_id_ << "]");
        }
        break;
    }

    case EKORDItemID::eCommandMoveDirect:
    {
        CLOG_INFO("Direct movement command is captured...");
        CLOG_INFO("Processing Request...");
        std::array<double, 7UL> q = a_controls.move_j_.getTargets();
        std::array<double, 7UL> qd = a_controls.move_j_.getTargetsD();
        std::array<double, 7UL> qdd = a_controls.move_j_.getTargetsDD();
        std::array<double, 7UL> torque = a_controls.move_j_.getTargetsTorque();

        sendTTCommandD(q, qd, qdd, torque);

        break;
    }

    case EKORDItemID::eCommandMoveVelocity:
    {
        if (a_recent)
        {
            //            CLOG_INFO("Velocity movement command is captured... (" << a_controls.capture_ts_.time_since_epoch().count()<< ")");
            //            CLOG_INFO("Processing Request...");
            DBG_COUT("Velocity movement command captured");
            std::array<double, 6UL> v = a_controls.move_l_.getTargets();
            long sync = a_controls.move_l_.getSyncValue();
            double period = a_controls.move_l_.getPeriod();
            double timeout = a_controls.move_l_.getTimeout();

            sendTTCommandV(v, sync, period, timeout);
        }
        break;
    }

    case EKORDItemID::eCommandSetFrame:
    {
        if (a_recent)
        {
            CLOG_INFO("Set Frame command is captured...");
            CLOG_INFO("Processing Request...");

            if (a_controls.crc_error_){
                CommandStatus new_stat = CommandStatus();
                new_stat.token = a_controls.time_stamp_token_;
                new_stat.error_code = 3; // CRC Error
                command_statuses_.addStatus(new_stat);
                return;
            }

            unsigned int exec_token_;
            auto result = sendTTCommandSetFrame(a_controls.frame_id_, a_controls.pose_, a_controls.ref_frame_, exec_token_);
            if (result.err_code_ != 0){ // not 0, reporting right away
                CommandStatus new_stat = CommandStatus();
                new_stat.token = a_controls.time_stamp_token_;
                new_stat.error_code = (result.err_code_ << 2) | 1; // RC API Error
                command_statuses_.addStatus(new_stat);
            }
            else{ // adding to pending
                PendingStatus new_stat = PendingStatus();
                new_stat.rc_token = exec_token_;
                new_stat.token = a_controls.time_stamp_token_;
                pending_statuses_.addStatus(new_stat);
            }
            
            CLOG_INFO("Got token: " << a_controls.time_stamp_token_);
            CLOG_INFO("Got error code: " << static_cast<int16_t>(result.err_code_));
        }

        break;
    }

    case EKORDItemID::eCommandSetLoad:
    {
        if (a_recent)
        {
            CLOG_INFO("Set Load command is captured...");
            CLOG_INFO("Processing Request...");

            if (a_controls.crc_error_){
                CommandStatus new_stat = CommandStatus();
                new_stat.token = a_controls.time_stamp_token_;
                new_stat.error_code = 3; // CRC Error
                command_statuses_.addStatus(new_stat);
                return;
            }

            unsigned int exec_token_;
            auto result = sendTTCommandSetLoad(a_controls.load_id_, a_controls.load_inertia_, a_controls.load_mass_, a_controls.load_cog_, exec_token_);
            if (result.err_code_ != 0){ // not 0, reporting right away
                CommandStatus new_stat = CommandStatus();
                new_stat.token = a_controls.time_stamp_token_;
                new_stat.error_code = (result.err_code_ << 2) | 1; // RC API Error
                command_statuses_.addStatus(new_stat);
            }
            else{ // adding to pending
                PendingStatus new_stat = PendingStatus();
                new_stat.rc_token = exec_token_;
                new_stat.token = a_controls.time_stamp_token_;
                pending_statuses_.addStatus(new_stat);
            }

            CLOG_INFO("Got token: " << a_controls.time_stamp_token_);
            CLOG_INFO("Got error code: " << static_cast<int16_t>(result.err_code_));
        }

        break;
    }

    case EKORDItemID::eCommandCleanAlarm:
    {
        if (a_recent)
        {
            CLOG_INFO("Clean Alarm command is captured...");
            CLOG_INFO("Processing Request...");

            unsigned int exec_token_;
            auto result = sendTTCommandCleanAlarm(a_controls.alarm_id_, exec_token_, stats_commands);
            if (result.err_code_ != 0){ // not 0, reporting right away
                CommandStatus new_stat = CommandStatus();
                new_stat.token = a_controls.time_stamp_token_;
                new_stat.error_code = (result.err_code_ << 2) | 1; // RC API Error
                command_statuses_.addStatus(new_stat);
            }
            else{
                if (a_controls.alarm_id_ != CBUN_EVENT){ // rc based commands
                    // adding to pending
                    PendingStatus new_stat = PendingStatus();
                    new_stat.rc_token = exec_token_;
                    new_stat.token = a_controls.time_stamp_token_;
                    new_stat.times_checked = 0;
                    pending_statuses_.addStatus(new_stat);
                    CLOG_INFO("Put rc_token token: " << new_stat.rc_token);
                }
                else{
                    // straigt to the statuses, without pending
                    CommandStatus new_stat = CommandStatus();
                    new_stat.token = a_controls.time_stamp_token_;
                    new_stat.error_code = (1 << 2) | 2; // Command error (6 = good)
                    command_statuses_.addStatus(new_stat);
                }
            }

            CLOG_INFO("Got token: " << a_controls.time_stamp_token_);
            CLOG_INFO("Got error code: " << static_cast<int16_t>(result.err_code_));
        }

        break;
    }

    default:
    {
        CLOG_INFO("[CBUN_KORD] Invalid id: " << static_cast<unsigned long>(a_controls.item_id_));
        break;
    }
    }
}

bool KORDReceiver::askForDashboardJsonToFile(){
    kr2rc_api::Controller::CmdUserConsentParams consent_params;
    std::string update_dash_cmd = "ctl/dashboard/refresh";
    consent_params.data_ = update_dash_cmd.c_str();
    consent_params.data_len_ = update_dash_cmd.length();
    program_api_->rc_api_->system_controller_->cmd_UserConsent(consent_params);
    return true;
}

// void KORDReceiver::syncRCData(kr2::kord::KR2KORDFrame_StatusJ_Res *a_jstate){
//     syncTick(0);
//     program_api_->rc_api_->spin();

//    const kr2rc_api::Model::ModelJoints *mj = program_api_->rc_api_->arm_model_->read_ModelJointsByDuid(kr2rc_api::Model::SysId::JOINT_POSITIONS);
//    if (mj){
//        a_jstate->setJointState(mj->values_.data());
//    }
//}

void KORDReceiver::syncTick(const timespec &a_sync_ts, CommandsMonitor& stats_commands)
{
    timespec next_tick = kr2rc_api_internals::Timex::nextTick(a_sync_ts);
    kr2rc_api_internals::Timex::nanosleepUntil(next_tick);
    // measuring wake Up stats
    auto diff_ = kr2rc_api_internals::Timex::nsecDiff(kr2rc_api_internals::Timex::now(), next_tick);
    stats_commands.processNewSystemJitter(diff_);
}

void KORDReceiver::syncTick(int a_ticks_to_future, CommandsMonitor& stats_commands, timespec a_tick_start)
{
    timespec ts = kr2rc_api_internals::Timex::now();
    timespec next_tick = kr2rc_api_internals::Timex::thisTick(ts, a_ticks_to_future);
    kr2rc_api_internals::Timex::nanosleepUntil(next_tick);

    auto diff_ = kr2rc_api_internals::Timex::nsecDiff(kr2rc_api_internals::Timex::now(), next_tick);
    //next_tick - now
    stats_commands.processNewSystemJitter(diff_);
    if (a_tick_start.tv_sec != 0 && a_tick_start.tv_nsec != 0)
    {
        // Calculate time skew
        double tick_overflow = kr2rc_api_internals::Timex::diff(a_tick_start, ts);
        if (tick_overflow > kord_period_){
            CLOG_WARN("Tick overflow detected: " << tick_overflow);
        }
    }


}


uint32_t KORDReceiver::systemAlarmState(CommandsMonitor& stats_commands){
    const int CONDITION_ID_OFFSET = 8;
    const int CNTXT_OFFSET = 4;

    uint32_t output = 0;

    kr2::kord::protocol::SystemEvent last_event;
    if (recent_soft_stop_event_.event_group_ != 0){
        last_event = recent_soft_stop_event_;
    }
    else{
        if (recent_safety_event_.event_group_ != 0){
            last_event = recent_safety_event_;
        }
    }
    switch (last_event.event_group_)
    {
        case EEventGroup::eSoftStopEvent:{
            if (last_event.event_id_ < 1000){ // invalid unknown id -> ignoring
                return 0;
            }
            output |= ESystemAlarmCategory::CAT_SOFT_STOP_EVENT;
            output |= last_event.event_id_ << CONDITION_ID_OFFSET;
            break;
        }
        case EEventGroup::eSafetyEvent:{
            if (last_event.event_id_ < 1000){ // invalid unknown id -> ignoring
                return 0;
            }
            output |= ESystemAlarmCategory::CAT_SAFETY_EVENT;
            // check for known ConditionID and only then return
            output |= last_event.event_id_ << CONDITION_ID_OFFSET; // EXTERNAL_PSTOP_ACTIVATED = 2001 can overflow to reserved bits, resulted in 233.
            break;
        }
        default:{
            if (stats_commands.getAlarmState()){ // CBunEvent (Stats) goes to Soft Stop
                output |= ESystemAlarmCategory::CAT_SOFT_STOP_EVENT;
                output |= ESoftStopEventConditionID::CBUN_KORD_BAD_CONN_QUALITY << CONDITION_ID_OFFSET;
            }
            else{
                // HW stat, no events
                if ((controller_model_.rc_hw_flags_ & EHWFlags::HW_STATUS_ERROR_BITS_SET) != 0){ // no hardware flags
                    output |= ESystemAlarmCategory::CAT_HW_STAT;
                }
                output |= controller_model_.rc_hw_flags_ << CONDITION_ID_OFFSET; // overflow possible
            }
            break;
        }
    }

    if (controller_model_.rc_safety_flags_ & ESafetyFlags::SAFETY_FLAG_SSTOP){
        output |= ESystemAlarmContext::CNTXT_SSTOP << CNTXT_OFFSET; // SoftStop
    }
    else if ((controller_model_.rc_safety_flags_ & ESafetyFlags::SAFETY_FLAG_PSTOP)
        || (controller_model_.rc_button_flags_ & EButtonFlags::BUTTONS_FLAG_PSTOP)){
        output |= ESystemAlarmContext::CNTXT_PSTOP << CNTXT_OFFSET; // ProtectiveStop
    }
    else if ((controller_model_.rc_safety_flags_ & ESafetyFlags::SAFETY_FLAG_ESTOP)
        || (controller_model_.rc_button_flags_ & EButtonFlags::BUTTONS_FLAG_ESTOP)){
        output |= ESystemAlarmContext::CNTXT_ESTOP << CNTXT_OFFSET; // EmergencyStop
    }
    else if ((controller_model_.rc_hw_flags_ & EHWFlags::HW_STATUS_ERROR_BITS_SET)
            || stats_commands.getAlarmState()){
        output |= ESystemAlarmContext::CNTXT_SYSERR << CNTXT_OFFSET; // SystemError
    }

    return output;
}

bool KORDReceiver::syncModel(CommandsMonitor& stats_commands)
{
    using namespace kr2rc_api;
    bool fetch_state = true;

    if (!program_api_->rc_api_)
    {
        CLOG_ERR("[KORD_CBUN] RC API ptr invalid");
    }

    program_api_->rc_api_->spin();
// timestamp
    fetch_state |= readModelJoints(arm_model_.joint_position_, Model::SysId::JOINT_POSITIONS);
    fetch_state |= readModelJoints(arm_model_.joint_speed_, Model::SysId::JOINT_SPEEDS);
    fetch_state |= readModelJoints(arm_model_.joint_acceleration_, Model::SysId::JOINT_ACCLERATIONS);
    fetch_state |= readModelJoints(arm_model_.joint_torque_, Model::SysId::JOINT_TORQUES);
//
    fetch_state |= readModelTF(arm_model_.tcp_model_, Model::SysId::FRAME_ROBOTX_TCP);
    fetch_state |= readModelTF(arm_model_.tcp_model_fwk_, Model::SysId::FRAME_ROBOT_FLANGE);
//
    const State::SysStat_RCState *ss_rcstate = program_api_->rc_api_->system_state_->read_SysStat_RCState();
    if (ss_rcstate)
    {
        controller_model_.rc_safety_flags_ = ss_rcstate->unsignedValue(State::SysStat_RCState::UITM_SAFETY_FLAGS);
        controller_model_.master_speed_ = ss_rcstate->floatValue(State::SysStat_RCState::FITM_MASTERSPEED);
        controller_model_.safety_mode_ = ss_rcstate->unsignedValue(State::SysStat_RCState::UITM_SAFETY_MODE);
        controller_model_.motion_flags_ = ss_rcstate->unsignedValue(State::SysStat_RCState::UITM_MOTION_FLAGS);
        controller_model_.rc_hw_flags_ = ss_rcstate->unsignedValue(State::SysStat_RCState::UITM_HW_STAT);
        controller_model_.rc_button_flags_ = ss_rcstate->unsignedValue(State::SysStat_RCState::UITM_SYSTEM_BUTTONS);
    }
//
    unsigned int j = 0;
    for (auto &js : arm_model_.sensors_)
    {
        fetch_state |= readJointSensors(js, j);
        j++;
    }
//
    // CPU State
    int counter = 0;
    std::vector <double> temps_;
    cpu_reader_.getRecentTemperatures(temps_);
    for (auto a_cpu_state: cpu_state_id_map_){
        controller_model_.cpu_state_[counter].value_ = temps_[counter];
        controller_model_.cpu_state_[counter].id_ = a_cpu_state;
        counter += 1;
    }
//
    // IOBoard State
    fetch_state |= readIOBoardSensors(arm_model_.ioboard_);
//
    // Frames
    counter = 0;
    for (auto a_frame: frame_id_map_){
        for (auto a_ref: frame_val_map_){
            fetch_state |= readModelTFWithRef(arm_model_.frames_[counter].pose_, a_frame.second, a_ref.second);
            arm_model_.frames_[counter].id_ = a_frame.first;
            arm_model_.frames_[counter].ref_value_id_ = a_ref.first;
            counter += 1;
            //CLOG_INFO(arm_model_.frames_[counter].id_ << " has pose[0] " << arm_model_.frames_[counter].pose_[0]);
        }
    }
//
    // Loads
    counter = 0;
    for (auto a_load: load_id_map_){
        fetch_state |= readModelLoadCoG(arm_model_.loads_[counter].cog_, a_load.second);
        fetch_state |= readModelLoadPose(arm_model_.loads_[counter].pose_, a_load.second);
        fetch_state |= readModelLoadInertia(arm_model_.loads_[counter].inertia_, a_load.second);
        fetch_state |= readModelLoadMass(arm_model_.loads_[counter].mass_, a_load.second);
        arm_model_.loads_[counter].id_ = a_load.first;
        //CLOG_INFO(a_load.first << " has cog[0] " << arm_model_.loads_[counter].cog_[0]);
        counter += 1;
    }
//
    //

    fetch_state |= readPeripherals();
//
    arm_model_.io_model_.setInputDigital(arm_model_.digital_inputs_);
    arm_model_.io_model_.setOutputDigital(arm_model_.digital_outputs_);
//
    // arm_model_.io_model_.setInputAnalog(arm_model_.analog_inputs_);
    // arm_model_.io_model_.setOutputAnalog(arm_model_.analog_outputs_);
    // Reading SystemEvent and updating *events* vector
    fetch_state |= readSystemEvents(controller_model_.system_events_);
//
    controller_model_.rc_system_alarm_state_ = systemAlarmState(stats_commands);
//
    return fetch_state;
}

template <class A, size_t B>
void KORDReceiver::sendTTCommand(const std::array<A, B> &a_target, const kr2rc_api::Trajectory::CmdTrajectoryParams &params)
{
    if (a_target.size() != B)
        return;

    if (a_target.size() == 6)
    {
        kr2rc_api::Pose target;
        target.setPRPY(a_target.data(), a_target.data() + 3); // todo: why +3?
        program_api_->rc_api_->arm_trajectory_->cmd_Follow(target, &params);
    }

    if (a_target.size() == 7)
    {
        kr2rc_api::JSVector target(a_target.data());
        program_api_->rc_api_->arm_trajectory_->cmd_Follow(target, &params);
    }
}

void KORDReceiver::sendTTCommandD(
    const std::array<double, 7UL> &a_q,
    const std::array<double, 7UL> &a_qd,
    const std::array<double, 7UL> &a_qdd,
    const std::array<double, 7UL> &a_torque)
{
    kr2rc_api::Drive::CmdDriveCtl cmd_drv_ctl;

    CLOG_INFO("[CBUN_KORD] sendTTCommandD");

    typedef kr2rc_api::Drive::CmdDirectJointCtl::EJointCtlModes JCTLM;

    // todo
    kr2rc_api::Drive::CmdDirectJointCtl::EJointCtlModes modes[7] = {
        JCTLM::ePositionCtl, JCTLM::ePositionCtl, JCTLM::ePositionCtl,
        JCTLM::ePositionCtl, JCTLM::ePositionCtl, JCTLM::ePositionCtl,
        JCTLM::ePositionCtl};

    kr2rc_api::JSVector pos(a_q.data());
    kr2rc_api::JSVector speed(a_qd.data());
    kr2rc_api::JSVector acc(a_qdd.data());
    kr2rc_api::JSVector torque(a_torque.data());

    kr2rc_api::Drive::CmdDirectJointCtl &dir_pos_ctl = cmd_drv_ctl
                                                           .asDirectControl()
                                                           .setJointModes(modes)
                                                           .withTargetPosition(pos)
                                                           .withTargetSpeed(speed)
                                                           .withTargetAcc(acc)
                                                           .withTargetTorque(torque)
                                                           .withTimeout(0.01);

    auto result = program_api_->rc_api_->arm_drive_->cmd_DriveCtl(dir_pos_ctl); // ignoring result...

    CLOG_INFO("[CBUN_KORD] sendTTCommandD DONE");
}

void KORDReceiver::sendTTCommandV(
    const std::array<double, 6UL> &a_v,
    const long a_sync_value,
    const double a_period,
    const double a_timeout)
{
    kr2rc_api::Drive::CmdDriveCtl cmd_drv_ctl;

    CLOG_INFO("[CBUN_KORD] sendTTCommandV");

    std::array<double, 3UL> a_trans_vel = {a_v[0], a_v[1], a_v[2]};
    std::array<double, 3UL> a_angl_vel = {a_v[3], a_v[4], a_v[5]};

    kr2rc_api::Vector vel(a_trans_vel.data());
    kr2rc_api::Vector omg(a_angl_vel.data());

    kr2rc_api::Drive::CmdDriveCtl drv_req;
    kr2rc_api::CmdResult result;
    // Handle zero parameters as velocity mode termination
    if (vel.Norm() == 0. && omg.Norm() == 0. && a_period == 0. && a_timeout == 0.)
    {
        result = program_api_->rc_api_->arm_drive_->cmd_DriveCtl(drv_req.asVelocityControl().terminate());
    }
    else
    {
        result = program_api_->rc_api_->arm_drive_->cmd_DriveCtl(
            drv_req
                .asVelocityControl()
                .withTargetVel(vel, omg)
                //.withSync(a_sync_value)
                .withSync(-1)
                .withPeriod(a_period)
                .withTimeout(a_timeout));
    }

    // Note: Ignoring result in this version

    CLOG_INFO("[CBUN_KORD] sendTTCommandV DONE");
}

kr2rc_api::CmdResult KORDReceiver::sendTTCommandSetFrame(
    unsigned int frame_id,
    const std::array<double, 6UL> &a_p, unsigned int ref_frame,
    unsigned int& exec_token_)
{
    kr2rc_api::Model::TF cmd_model_TF;
    kr2rc_api::Frame frame;
    CLOG_INFO("[CBUN_KORD] sendTTCommandFrame");

    kr2rc_api::Vector pos(a_p[0], a_p[1], a_p[2]);

    kr2rc_api::Rotation rot = kr2rc_api::Rotation::RPY(a_p[3], a_p[4], a_p[5]);

    frame.M_ = rot;
    frame.p_ = pos;

    cmd_model_TF.pose_ = frame;

    // frame ID
    if ((frame_id_map_.find(frame_id) == frame_id_map_.end()) || (frame_val_map_.find(ref_frame) == frame_id_map_.end())){
        return kr2rc_api::CmdResult{404, -1};
    }
    cmd_model_TF.frame_id_ = frame_id_map_[frame_id];
    cmd_model_TF.ref_frame_id_ = frame_val_map_[ref_frame];

    kr2rc_api::Model::CmdSetSystemFrameParams token_param = kr2rc_api::Model::CmdSetSystemFrameParams();
    token_param.rc_token_id_ = last_token_;

    // CLOG_INFO("!!Before new command");
    // CLOG_INFO("!SysStat Start");
    // long sys_state_n = program_api_->rc_api_->system_state_->read_SysStat_NReceipts();
    // CLOG_INFO("SysStat Receipts size: " << sys_state_n);
    // for (int i = 0; i < sys_state_n; i++){
    //     auto new_one = program_api_->rc_api_->system_state_->read_SysStat_Receipt(i);
    //     CLOG_INFO("SysStat with token: " << new_one->value(kr2rc_api::State::SysStat_Receipt::EItem::UITM_TOKEN) << " with success: " << new_one->value(kr2rc_api::State::SysStat_Receipt::EItem::UITM_SUCCESS));
    // }
    // CLOG_INFO("!SysStat End");

    auto res = program_api_->rc_api_->arm_model_->cmd_SetSystemFrame(&cmd_model_TF, &token_param);

    CLOG_INFO("!!! Executed with token: " << last_token_);
    exec_token_ = last_token_;

    last_token_ = (last_token_ + 1) % TOKEN_MAX_;

    CLOG_INFO("[CBUN_KORD] sendTTCommandFrame DONE");

    return res;
}

kr2rc_api::CmdResult KORDReceiver::sendTTCommandSetLoad(
    unsigned int a_load_id,
    const std::array<double, 6UL> &a_inertia,
    double a_mass,
    const std::array<double, 3UL> &a_cog,
    unsigned int& exec_token_)
{
    kr2rc_api::Load::AttachedLoad cmd_model_Load;
    CLOG_INFO("[CBUN_KORD] sendTTCommandLoad");

    // load ID
    switch (a_load_id)
    {
    case LOAD1:
    {
        cmd_model_Load.load_id_ = kr2rc_api::Load::SysId::LOAD_TOOL;
        break;
    }
    case LOAD2:
    {
        cmd_model_Load.load_id_ = kr2rc_api::Load::SysId::LOAD_PAYLOAD;
        break;
    }
    }
    const double inertia[6] = {a_inertia[0], a_inertia[1], a_inertia[2],
                               a_inertia[3], a_inertia[4], a_inertia[5]};
    const double mass = a_mass;
    const double com[3] = {a_cog[0], a_cog[1], a_cog[2]};
    kr2rc_api::Inertia in(mass, com, inertia);
    cmd_model_Load.imx_ = in;

    kr2rc_api::Load::CmdSetLoadParams token_param = kr2rc_api::Load::CmdSetLoadParams();
    token_param.rc_token_id_ = last_token_;

    auto res =  program_api_->rc_api_->arm_load_->cmd_Set(&cmd_model_Load, &token_param);

    CLOG_INFO("!!! Executed with token: " << last_token_);
    exec_token_ = last_token_;

    last_token_ = (last_token_ + 1) % TOKEN_MAX_;

    CLOG_INFO("[CBUN_KORD] sendTTCommandLoad DONE");

    return res;
}

kr2rc_api::CmdResult KORDReceiver::sendTTCommandCleanAlarm(
    unsigned int a_alarm_id, unsigned int& exec_token_, CommandsMonitor& stats_commands)
{
    CLOG_INFO("[CBUN_KORD] sendTTCommandCleanAlarm");

    std::string rc_command;

    switch (a_alarm_id)
    {
    case HALT:
        rc_command = "ctl/clear-halt";
        break;
    case SUSPENDED:
    case PAUSED:
        rc_command = "ctl/unsuspend";
        break;
    case INIT:
        rc_command = "init/discovery/consent";
        break;
    case CBUN_EVENT:
        // not proceeding to the rc part
        stats_commands.cleanAlarmState();

        exec_token_ = last_token_;
        last_token_ = (last_token_ + 1) % TOKEN_MAX_;

        auto res = kr2rc_api::CmdResult();
        res.err_code_ = 0;
        return res;
    }
    

    kr2rc_api::Controller::CmdUserConsentParams consent_params;

    consent_params.token_ = last_token_;

    consent_params.data_ = rc_command.c_str();
    consent_params.data_len_ = rc_command.length();

    auto res =  program_api_->rc_api_->system_controller_->cmd_UserConsent(consent_params);

    CLOG_INFO("!!! Executed with token: " << last_token_);
    exec_token_ = last_token_;

    last_token_ = (last_token_ + 1) % TOKEN_MAX_;

    CLOG_INFO("[CBUN_KORD] sendTTCommandCleanAlarm DONE");

    return res;
}

void KORDReceiver::sendSetIOCommand(const RobotControl &a_controls)
{
    CLOG_INFO("[CBUN_KORD] sendSetIOCommand");

    kr2rc_api::IOData::CmdTXGPIOParams cmd_params;
    kr2rc_api::CmdResult result;
    // Collecting Relay+Digital+(TB)
    // size                                   first: 4            +      next: DO 8         +  next:  Tool IO 4
    size_t num_outputs_relay_digitals_tio = num_outputs_relay_iob + num_outputs_digital_iob + num_outputs_digital_tio;

    for (size_t i = 0; i < num_outputs_relay_digitals_tio; i++)
    {
        kr2rc_api::DUID duid = output_rdt_duid_[i];
        bool is_set = std::bitset<32>(a_controls.mask_).test(i);
        // unsigned int config_id_ = 0;

        // find the right port by duid and set value if set
        for (auto &do_ : arm_model_.digital_outputs_)
        {
            if (do_.gpio_id_ == duid)
            {
                if (is_set)
                {
                    do_.value_ = a_controls.v_;

                    // if it is TOOL Board --> configuration must be 24V
                    bool is_tb = (std::find(output_tb_duid_.begin(), output_tb_duid_.end(), do_.gpio_id_) != output_tb_duid_.end());
                    if (is_tb)
                    {
                        // !
                        // TODO AK:  it has to be like this, because if you set it within array it will not reflect [not expected way]
                        kr2rc_api::IOData::GPIOInt64 digital{do_.gpio_id_, a_controls.v_, KR2_TB_CONF_OUT_DIG_VOLTAGE_24V};
                        result = program_api_->rc_api_->iob_data_->cmd_TX_GPIO(cmd_params, nullptr, 0, &digital, 1, NULL, 0);

                        // ?
                        // do_.config_id_ = KR2_TB_CONF_OUT_DIG_VOLTAGE_24V; // default 24V // [expected way]
                    }
                }
            }
        }
    }

    result = program_api_->rc_api_->iob_data_->cmd_TX_GPIO(cmd_params, nullptr, 0, arm_model_.digital_outputs_.data(), arm_model_.digital_outputs_.size(), NULL, 0);

    arm_model_.io_model_.setOutputDigital(arm_model_.digital_outputs_);

    CLOG_INFO("[CBUN_KORD] sendSetIOCommand DONE");
}

kr2rc_api::CmdResult KORDReceiver::sendDJC_CommandFW(unsigned int a_jidx, kr2::kord::protocol::EJointFirmwareCommand a_command, unsigned int& exec_token_)
{
    kr2rc_api::Drive::CmdDriveCtl cmd_drv_ctl;
    std::string scmd;
    switch (a_command)
    {
    case kr2::kord::protocol::EJointFirmwareCommand::eBrakeClear:
    {
        scmd = "clear";
        break;
    }
    case kr2::kord::protocol::EJointFirmwareCommand::eBrakeDisengage:
    {
        scmd = "release";
        break;
    }
    case kr2::kord::protocol::EJointFirmwareCommand::eBrakeEngage:
    {
        scmd = "lock";
        break;
    }
    default:
    {
        scmd = "clear";
        break;
    }
    }
    kr2rc_api::Model::CmdSetSystemFrameParams token_param = kr2rc_api::Model::CmdSetSystemFrameParams();
    token_param.rc_token_id_ = last_token_; // the same logic as in SetFrame/SetLoad

    kr2rc_api::Drive::CmdDirectJointCtl &brk_release = cmd_drv_ctl
                                                           .asDirectControl()
                                                           .withJointFWCommand(a_jidx, scmd.c_str(), scmd.size())
                                                           .withTimeout(0.01);
    brk_release.withToken(token_param.rc_token_id_);

    auto result = program_api_->rc_api_->arm_drive_->cmd_DriveCtl(brk_release);

    CLOG_INFO("!!! Executed with token: " << last_token_);
    exec_token_ = last_token_;

    last_token_ = (last_token_ + 1) % TOKEN_MAX_;
    return result;
}

void KORDReceiver::processFirmwareCommands(RobotControl &a_controls)
{
    using namespace std::chrono;
    milliseconds elapsed = duration_cast<milliseconds>(std::chrono::high_resolution_clock::now() - a_controls.last_fw_cmd_processed_);

    //
    if (a_controls.failed_by_not_completed != 1){
        CommandStatus new_stat = CommandStatus();
        new_stat.token = a_controls.failed_by_not_completed;
        new_stat.error_code = 0; // Too many commands at once, sending back the one that just came overlaping
        command_statuses_.addStatus(new_stat);
        a_controls.failed_by_not_completed = 1;
    }

    if (elapsed > CMD_FW_DELAY && !a_controls.jnt_fw_commands_.empty())
    {
        // for getting the response
        if (a_controls.crc_error_){
            CommandStatus new_stat = CommandStatus();
            new_stat.token = a_controls.time_stamp_token_;
            new_stat.error_code = 3; // CRC Error
            command_statuses_.addStatus(new_stat);

            a_controls.jnt_fw_commands_.clear();
            return;
        }

        unsigned int exec_token_;

        a_controls.last_fw_cmd_processed_ = std::chrono::high_resolution_clock::now();
        auto result = sendDJC_CommandFW(a_controls.jnt_fw_commands_.back().first, a_controls.jnt_fw_commands_.back().second, exec_token_);

        if (result.err_code_ != 0){ // not 0, reporting right away
            CommandStatus new_stat = CommandStatus();
            new_stat.token = a_controls.time_stamp_token_;
            new_stat.error_code = (result.err_code_ << 2) | 1; // RC API Error
            command_statuses_.addStatus(new_stat);
        }
        else{ // adding to pending
            // because every brake is activated separately we should have 7 results for each API command
            // with the same token
            PendingStatus new_stat = PendingStatus();
            new_stat.rc_token = exec_token_;
            new_stat.token = a_controls.time_stamp_token_;
            new_stat.composed = true;
            new_stat.compos_left = a_controls.jnt_fw_commands_.size() - 1; // because this one is already considered
            new_stat.compos_total = a_controls.composed_dependent_;
            if (new_stat.compos_left + 1 == a_controls.composed_dependent_){
                accum_pending_results[new_stat.token] = 0;
                accum_pending_result = 0; // first of the series => RESET
            }
            pending_statuses_.addStatus(new_stat);
        }
        CLOG_INFO("Got joint brake" << '\n');
        CLOG_INFO("Got token: " << a_controls.time_stamp_token_);
        CLOG_INFO("Got error code: " << static_cast<int16_t>(result.err_code_));

        a_controls.jnt_fw_commands_.pop_back();
    }
}

bool KORDReceiver::readModelTF(std::vector<double> &a_out, kr2rc_api::DUID a_duid)
{
    const kr2rc_api::Model::TF *tcp = program_api_->rc_api_->arm_model_->read_TransformationByDuid(a_duid);
    if (!tcp)
        return false;
    if (!tcp->pose_.p_.valid())
        return false;
    if (!tcp->pose_.M_.valid())
        return false;

    a_out.clear();
    double r, p, y;
    a_out.insert(a_out.begin(), tcp->pose_.p_.data(), tcp->pose_.p_.data() + 3);
    tcp->pose_.M_.getRPY(r, p, y);
    a_out.push_back(r);
    a_out.push_back(p);
    a_out.push_back(y);

    return true;
}

bool KORDReceiver::readModelTFWithRef(std::vector<double> &a_out, kr2rc_api::DUID a_duid, kr2rc_api::DUID a_ref_duid)
{
    const kr2rc_api::Model::TF *tcp = program_api_->rc_api_->arm_model_->read_TransformationByDuid(a_duid, a_ref_duid);
    if (!tcp)
        return false;
    if (!tcp->pose_.p_.valid())
        return false;
    if (!tcp->pose_.M_.valid())
        return false;

    a_out.clear();
    double r, p, y;
    a_out.insert(a_out.begin(), tcp->pose_.p_.data(), tcp->pose_.p_.data() + 3);
    tcp->pose_.M_.getRPY(r, p, y);
    a_out.push_back(r);
    a_out.push_back(p);
    a_out.push_back(y);

    // CLOG_INFO(a_out[0]);

    return true;
}

bool KORDReceiver::readModelLoadCoG(std::vector<double> &a_out, kr2rc_api::DUID a_duid)
{
    const kr2rc_api::Load::AttachedLoad *load = program_api_->rc_api_->arm_load_->read_AttachedLoadByDuid(a_duid);
    if (!load)
        return false;
    kr2rc_api::Vector com = load->imx_.centreOfMass();
    if (!com.valid())
        return false;

    a_out.clear();
    a_out.insert(a_out.begin(), com.data(), com.data() + 3);

    return true;
}

bool KORDReceiver::readModelLoadPose(std::vector<double> &a_out, kr2rc_api::DUID a_duid)
{
    const kr2rc_api::Load::AttachedLoad *load = program_api_->rc_api_->arm_load_->read_AttachedLoadByDuid(a_duid);
    if (!load)
        return false;
    kr2rc_api::Pose pose = load->load_frame_pose_;
    if (!pose.valid())
        return false;

    a_out.clear();
    double r, p, y;
    a_out.insert(a_out.begin(), pose.frame_.p_.data(), pose.frame_.p_.data() + 3);
    pose.frame_.M_.getRPY(r, p, y);
    a_out.push_back(r);
    a_out.push_back(p);
    a_out.push_back(y);

    return true;
}

bool KORDReceiver::readModelLoadInertia(std::vector<double> &a_out, kr2rc_api::DUID a_duid)
{
    const kr2rc_api::Load::AttachedLoad *load = program_api_->rc_api_->arm_load_->read_AttachedLoadByDuid(a_duid);
    if (!load)
        return false;
    kr2rc_api::Inertia inertia = load->imx_;

    a_out.clear();
    a_out.push_back(inertia.ixx());
    a_out.push_back(inertia.ixy());
    a_out.push_back(inertia.ixz());
    a_out.push_back(inertia.iyy());
    a_out.push_back(inertia.iyz());
    a_out.push_back(inertia.izz());

    return true;
}

bool KORDReceiver::readModelLoadMass(double &a_out, kr2rc_api::DUID a_duid)
{
    const kr2rc_api::Load::AttachedLoad *load = program_api_->rc_api_->arm_load_->read_AttachedLoadByDuid(a_duid);
    if (!load)
        return false;
    kr2rc_api::Inertia inertia = load->imx_;

    a_out = inertia.mass();

    return true;
}

bool KORDReceiver::readModelJoints(std::vector<double> &a_out, kr2rc_api::DUID a_duid)
{
    const kr2rc_api::Model::ModelJoints *mj = program_api_->rc_api_->arm_model_->read_ModelJointsByDuid(a_duid);
    if (!mj)
        return false;
    if (!mj->values_.valid())
        return false;

    a_out.clear();
    a_out.insert(a_out.begin(), mj->values_.data(), mj->values_.data() + 7);

    return true;
}

bool KORDReceiver::readJointSensors(JointSensorData &a_sensors, unsigned int a_idx)
{
    if (a_idx < 0 || a_idx > 6)
    {
        return false;
    }

    const kr2rc_api::Sensors::JointBoard *jb_sens = program_api_->rc_api_->arm_sensors_->read_SingleJoint(a_idx);

    if (!jb_sens)
    {
        return false;
    }

    a_sensors.status_bits_ = jb_sens->unsignedValue(kr2rc_api::Sensors::JointBoard::UITM_STATUS_BITS);
    a_sensors.error_bits_ = jb_sens->unsignedValue(kr2rc_api::Sensors::JointBoard::UITM_ERROR_BITS);

    a_sensors.temperature_[0] = jb_sens->floatValue(kr2rc_api::Sensors::JointBoard::FITM_TEMP_BOARD);
    a_sensors.temperature_[1] = jb_sens->floatValue(kr2rc_api::Sensors::JointBoard::FITM_TEMP_JOINTENC);
    a_sensors.temperature_[2] = jb_sens->floatValue(kr2rc_api::Sensors::JointBoard::FITM_TEMP_ROTORENC);

    a_sensors.current_ = jb_sens->floatValue(kr2rc_api::Sensors::JointBoard::FITM_STATOR_CURENT);
    a_sensors.voltage_ = jb_sens->floatValue(kr2rc_api::Sensors::JointBoard::FITM_STATOR_VOLTAGE);
    a_sensors.torque_ = jb_sens->floatValue(kr2rc_api::Sensors::JointBoard::FITM_TORQUE);

    return true;
}

bool KORDReceiver::readIOBoardSensors(IOBoardData &a_ioboard)
{
    // CLOG_INFO("Temperature start");
    if (!program_api_->rc_api_->iob_device_)
    {
        // CLOG_INFO("NO DEVICE");
        return false;
    }
    int duid_iob_temp_cabinet = 0x00108A03;
    if (!program_api_->rc_api_->iob_device_->read_DevValueByDuid(duid_iob_temp_cabinet))
    {
        // CLOG_INFO("NO OFFSET");
        return false;
    }
    a_ioboard.temperature_ = program_api_->rc_api_->iob_device_->read_DevValueByDuid(duid_iob_temp_cabinet)->floatValue(0.0);
    // CLOG_INFO("DONE");
    return true;
}

bool KORDReceiver::readPeripherals()
{
    bool ok = true;

    ok &= readDigital(kr2rc_api::IOData::IOTARGET_RXDATA, arm_model_.digital_inputs_);
    ok &= readAnalog(kr2rc_api::IOData::IOTARGET_RXDATA, arm_model_.analog_inputs_);

    ok &= readDigital(kr2rc_api::IOData::IOTARGET_TXDATA, arm_model_.digital_outputs_);
    ok &= readAnalog(kr2rc_api::IOData::IOTARGET_TXDATA, arm_model_.analog_outputs_);

    return ok;
}

bool KORDReceiver::readDigital(long a_dir, std::vector<kr2rc_api::IOData::GPIOInt64> &a_out)
{
    long direction_input = a_dir;
    unsigned int nitems = program_api_->rc_api_->io_data_->read_N_GPIOInt64(direction_input);

    if (a_out.size() != nitems)
    {
        a_out.resize(nitems);
    }

    unsigned int read = program_api_->rc_api_->io_data_->read_all_GPIOInt64(a_out.data(), a_out.size(), direction_input);

    return read == nitems;
}

bool KORDReceiver::readAnalog(long a_dir, std::vector<kr2rc_api::IOData::GPIOFloat> &a_out)
{
    long direction_input = a_dir;
    unsigned int nitems = program_api_->rc_api_->io_data_->read_N_GPIOFloat(direction_input);

    if (a_out.size() != nitems)
    {
        a_out.resize(nitems);
    }

    unsigned int read = program_api_->rc_api_->io_data_->read_all_GPIOFloat(a_out.data(), a_out.size(), direction_input);

    return read == nitems;
}

bool KORDReceiver::readSystemEvents(SystemEvents &events)
{

    // Currently only the even ids will be transferred. The details about the
    // violations are not being transferred since the buffer is too short. The
    // values and thresholds will become part of the information after we can
    // utilize the request and response schema.

    //
    // Reading Safety Events
    //
    long n_safety_event = program_api_->rc_api_->system_state_->read_SysStat_NSafetyEvents();

    SystemEvent system_event;
    for (unsigned int i = 0; i < n_safety_event; i++)
    {
        system_event.reset();
        const kr2rc_api::State::SysStat_SafetyEvent *safety_event;
        safety_event = program_api_->rc_api_->system_state_->read_SysStat_SafetyEvent(i);
        if (!safety_event)
        {
            // Consider return false - this means there are no data for an event
            continue;
        }
        timespec ts = safety_event->timeStamp();
        system_event.timestamp_ = kr2::utils::timespecToEpochNanosec(ts);
        system_event.event_group_ = EEventGroup::eSafetyEvent;
        system_event.event_id_ = safety_event->unsignedValue(kr2rc_api::State::SysStat_SafetyEvent::EItem::UITM_CONDITION);
        events.addEvent(system_event);
    }
    recent_safety_event_ = system_event;
    //
    // Reading SoftStop Events
    //
    long n_soft_stop = program_api_->rc_api_->system_state_->read_SysStat_NSStopEvents();
    // std::cout << "N Soft Stop Event: " << n_soft_stop << "\n";

    system_event.reset();
    for (unsigned int i = 0; i < n_soft_stop; i++)
    {
        system_event.reset();
        const kr2rc_api::State::SysStat_SoftStopEvent *soft_stop_event;
        soft_stop_event = program_api_->rc_api_->system_state_->read_sysStat_StopEvent(i);
        if (!soft_stop_event)
        {
            // Consider return false - this means there are no data for an event
            continue;
        }
        timespec ts = soft_stop_event->timeStamp();
        system_event.timestamp_ = kr2::utils::timespecToEpochNanosec(ts);
        system_event.event_group_ = EEventGroup::eSoftStopEvent;
        system_event.event_id_ = soft_stop_event->unsignedValue(kr2rc_api::State::SysStat_SoftStopEvent::EItem::UITM_CONDITION);
        // convert to system event
        events.addEvent(system_event);
    }
    recent_soft_stop_event_ = system_event;

    return true;
}

class UDPKord::Internals
{
public:
    Internals(boost::shared_ptr<kr2_program_api::ProgramInterface> a_api)
        : udp_rcv_(new KORDReceiver(a_api))
    {
    }

    ~Internals() {}

    bool processActivationParams(const boost::property_tree::ptree &a_param_tree)
    {
        kr2_bundle_api::ArgProviderXml arg_provider(a_param_tree);

        if (arg_provider.getArgCount() != 4)
        {
            CLOG_ERR("[CBUN/UDP_KORD] Unexpected param count=" << arg_provider.getArgCount());
            return false;
        }

        activation_params_.ifname = "ethnet";
        activation_params_.port = arg_provider.getInt(0);
        activation_params_.session_id = arg_provider.getInt(1);
        activation_params_.sync_w_rc = true;
        activation_params_.use_realtime_ = arg_provider.getBool(2);
        activation_params_.realtime_prio_ = arg_provider.getInt(3);

        return true;
    }

public:
    boost::shared_ptr<KORDReceiver> udp_rcv_;

    struct
    {
        int port;
        int session_id;
        std::string ifname;
        bool sync_w_rc;
        bool use_realtime_;
        int realtime_prio_;
        std::string config_file_path;
        int kord_period;
    } activation_params_;
};

UDPKord::UDPKord(boost::shared_ptr<kr2_program_api::ProgramInterface> a_api,
                 const boost::property_tree::ptree &a_xml_bundle_node)
    : kr2_bundle_api::CustomDevice(a_api, a_xml_bundle_node),
      his_(new Internals(a_api))
{
    //                                             host      username      path
    REGISTER_RPC(&UDPKord::sshExportLogs, this, ARG_STR(0), ARG_STR(1), ARG_STR(2));
    //                                               fname
    REGISTER_RPC(&UDPKord::sshExportLogsIni, this, ARG_STR(0));
}

UDPKord::~UDPKord()
{
}

int kswx_kr2_kord::UDPKord::onCreate()
{
    if (activation_tree_)
    {
        activate(*activation_tree_);
    }

    CLOG_INFO("\nKassowRobots KORD CBun\n");
    CLOG_INFO("    version: " << GIT_VERSION);
    CLOG_INFO("        sha: " << GIT_SHA1 << "\n");

    return 0;
}

int kswx_kr2_kord::UDPKord::onDestroy()
{
    return 0;
}

int kswx_kr2_kord::UDPKord::onBind()
{
    return 0;
}

int kswx_kr2_kord::UDPKord::onUnbind()
{
    return 0;
}

kr2_program_api::CmdResult<> kswx_kr2_kord::UDPKord::onActivate(const boost::property_tree::ptree &param_tree)
{
    if (!his_->processActivationParams(param_tree))
    {
        CBUN_PCALL_RET_ERROR(-1, "Invalid activation params");
    }
    //     his_->udp_rcv_->setPort(port);
    //     his_->udp_rcv_->setPortInterface(ifname);

    // NOTE check for errors
    CLOG_INFO("[KORD_CBUN] Start the rcv/trx threads");

    // following parameters now predefined, in the future probably difined by the user
    his_->activation_params_.config_file_path = "/home/nucky/kr2.root/Workspace/Files/";

    his_->udp_rcv_->start(
        his_->activation_params_.port,
        his_->activation_params_.session_id,
        his_->activation_params_.sync_w_rc,
        his_->activation_params_.use_realtime_,
        his_->activation_params_.realtime_prio_,
        his_->activation_params_.config_file_path
        );
    CLOG_INFO("[KORD_CBUN] Communication started.");

    CBUN_PCALL_RET_OK;
}

kr2_program_api::CmdResult<> kswx_kr2_kord::UDPKord::onDeactivate()
{    
    // NOTE check for errors
    his_->udp_rcv_->stop();
    CBUN_PCALL_RET_OK;
}

kr2_program_api::CmdResult<> kswx_kr2_kord::UDPKord::onMount(const boost::property_tree::ptree &param_tree)
{
    CBUN_PCALL_RET_OK;
}

kr2_program_api::CmdResult<> kswx_kr2_kord::UDPKord::onUnmount()
{
    CBUN_PCALL_RET_OK;
}

CBUN_PCALL UDPKord::sshExportLogs(const std::string &host, const std::string &username, const std::string &path)
{
    return uploadLogs(host, username, path);
}

// [KORD]
// username = alexander
// log_path = ~/kassow_logs_ini_file [optional]
// host     = 192.168.X.Y            [optional]
CBUN_PCALL UDPKord::sshExportLogsIni(const std::string &fname)
{
    return baseSshExportLogsIni(fname);
}
